(**************************************************************************)
(*  ProuveParser: parser frontend for the PROUVÉ language                *)
(*                                                                        *)
(*  Copyright (C) 2004-2009 Ralf Treinen                                  *)
(*                                                                        *)
(*  This library is free software; you can redistribute it and/or         *)
(*  modify it under the terms of the GNU Lesser General Public            *)
(*  License as published by the Free Software Foundation; either          *)
(*  version 2.1 of the License, or (at your option) any later version.    *)
(*                                                                        *)
(*  This library is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *)
(*  Lesser General Public License for more details.                       *)
(*                                                                        *)
(*  You should have received a copy of the GNU Lesser General Public      *)
(*  License along with this library; if not, write to the Free Software   *)
(*  Foundation, Inc., 59 Temple Place, Suite 330,                         *)
(*  Boston, MA  02111-1307  USA                                           *)
(*                                                                        *)
(*  Ralf Treinen <treinen@pps.jussieu.fr>                                 *)
(*  Laboratoire Preuves, Programmes et Systèmes (PPS),                   *)
(*  Université Paris Diderot, Paris, France                              *)
(*  http://www.pps.jussieu.fr/~treinen                                    *)
(*                                                                        *)
(*  This software was mainly written while the author was with the        *)
(*  Laboratoire Spécification et Vérification (LSV),                    *)
(*  ENS de Cachan, Cachan, France                                         *)
(**************************************************************************)

open Expressions
open Types
open Declarations

type axioms = declaration list * expression list

(* (is_axiom a) checks that a is an axiom, that is an equality between *)
(* two functional expressions *)
let is_axiom = function
    Infix(_,"=",_,_) -> true
  | _ -> false

let is_c_axiom = function
  | Infix(_,"=",
	  Apply(_,fl,[Ident(_,vl1);Ident(_,vl2)]),
	  Apply(_,fr,[Ident(_,vr1);Ident(_,vr2)])) ->
      if fl = fr && vl1 <> vl2 && vl1 = vr2 && vl2 = vr1
      then Some fl
      else None
  | _ -> None

let is_a_axiom = function 
  | Infix(_,"=",
	  Apply(_,fl1,[
		  Ident(_,vl1);
		  Apply(_,fl2,[
			  Ident(_,vl2);
			  Ident(_,vl3)])]),
	  Apply(_,fr1,[
		  Apply(_,fr2,[
			  Ident(_,vr1);
			  Ident(_,vr2)]);
		  Ident(_,vr3)]))
  | Infix(_,"=",
	  Apply(_,fr1,[
		  Apply(_,fr2,[
			  Ident(_,vr1);
			  Ident(_,vr2)]);
		  Ident(_,vr3)]),
	  Apply(_,fl1,[
		  Ident(_,vl1);
		  Apply(_,fl2,[
			  Ident(_,vl2);
			  Ident(_,vl3)])]))
    ->
      if fl1 = fl2 && fr1 = fr2 && fl1 = fr1
	&& vl1 <> vl2 && vl2 <> vl3 && vl1 <> vl3
	&& vl1 = vr1 && vl2 = vr2 && vl3 = vr3
      then Some fl1
      else None
  | _ -> None
      
exception Exc_axioms of string
let type_axioms sign (declarations,expressions) =
  begin
    (* check that all the declarations are well-formed *)
    check_declarations declarations;

    (* check that all variable declarations are logical *)
    List.iter
      (function (x,(k,_)) -> match k with
	   Mutable -> raise
	     (Exc_axioms ("variable \"" ^ x ^ "\" declared " 
			  ^ "as mutable in axiom section"))
	 | Logical -> ()
	 | ProcessKind -> failwith "this cannot happen: processkind in axiom")
      declarations;

    (* check that both sides of the equations are well-typed *)
    (* and have the same type.                               *)
    let expressions_typed =
      List.map
	(function expr ->	match expr with
	     Infix(info,"=",left,right) ->
	       let left_typed = 
		 try 
		   type_expr sign declarations [] [] left
		 with
		     Exc_typeexpression s -> raise
		       (Exc_axioms
			  ("expression \"" ^
			     (string_of_expression false left) ^
			     "\" cannot be typed because " ^ s))
	       and right_typed =
		 try
		   type_expr sign declarations [] [] right
		 with
		     Exc_typeexpression s -> raise
		       (Exc_axioms
			  ("expression \"" ^
			     (string_of_expression false right) ^
			     "\" cannot be typed because " ^ s)) 
	       in
	       let left_type = extract_type_expr left_typed
	       and right_type = extract_type_expr right_typed
	       in
		 if (has_inf_type left_type  right_type)
		 then Infix(
		   {info with exp_type = Typeconst "bool"},
		   "=",
		   left_typed,
		   right_typed)
		 else (raise
			 (Exc_axioms
			    ("axiomatic equation between expressions "
			     ^ "of inconsistentg type:\n" ^
			     "left side \"" ^
			     (string_of_expression false left) ^
			     "\" has type " ^
			     (string_of_type_expression left_type) ^ 
			     "\nright side \"" ^
			     (string_of_expression false right) ^
			     "\" has type " ^
			     (string_of_type_expression right_type ))))
	   | _ -> failwith "this cannot happen: non-equational axiom") 
	expressions
    in
      (declarations, expressions_typed)
  end
    
