(**************************************************************************)
(*  ProuveParser: parser frontend for the PROUVÉ language                *)
(*                                                                        *)
(*  Copyright (C) 2004-2009 Ralf Treinen                                  *)
(*                                                                        *)
(*  This library is free software; you can redistribute it and/or         *)
(*  modify it under the terms of the GNU Lesser General Public            *)
(*  License as published by the Free Software Foundation; either          *)
(*  version 2.1 of the License, or (at your option) any later version.    *)
(*                                                                        *)
(*  This library is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *)
(*  Lesser General Public License for more details.                       *)
(*                                                                        *)
(*  You should have received a copy of the GNU Lesser General Public      *)
(*  License along with this library; if not, write to the Free Software   *)
(*  Foundation, Inc., 59 Temple Place, Suite 330,                         *)
(*  Boston, MA  02111-1307  USA                                           *)
(*                                                                        *)
(*  Ralf Treinen <treinen@pps.jussieu.fr>                                 *)
(*  Laboratoire Preuves, Programmes et Systèmes (PPS),                   *)
(*  Université Paris Diderot, Paris, France                              *)
(*  http://www.pps.jussieu.fr/~treinen                                    *)
(*                                                                        *)
(*  This software was mainly written while the author was with the        *)
(*  Laboratoire Spécification et Vérification (LSV),                    *)
(*  ENS de Cachan, Cachan, France                                         *)
(**************************************************************************)

open Types

type declarationkind =
    Logical
  | Mutable
  | ProcessKind

type declaration = 
    string * (declarationkind * type_expression)

type sk_declaration =
    string * ( type_expression list * declarationkind * type_expression)

let get_kind_decl varname env = fst(List.assoc varname env)
let get_type_decl varname env = snd(List.assoc varname env)

(**** renaming of Variables **********************************************)

let rename_decl alpha (varname,blurb) =
  ((try List.assoc varname alpha with Not_found -> varname), blurb)
let rename_dlist alpha = List.map (rename_decl alpha) 

(**** Conversion to Strings **********************************************)
let string_of_declarationkind = function
    Logical -> ""
  | Mutable -> "mutable"
  | ProcessKind -> "process"

let string_of_declaration (varname, (declkind, decltyp)) =
  varname ^ ": " ^ (string_of_declarationkind declkind) ^
    " " ^ (string_of_type_expression decltyp)

let string_of_sk_declaration (varname, (argtypes, declkind, decltyp)) =
  varname ^ 
    (match argtypes with
	 [] -> ""
       | _ -> "(" ^
	   (Utils.string_of_string_list ", "
	      (List.map string_of_type_expression argtypes)) ^
	     ")") ^
    " : " ^ (string_of_declarationkind declkind) ^
    " " ^ (string_of_type_expression decltyp)

let string_of_parlist dl = 
  Utils.string_of_string_list "; "
    (List.map string_of_declaration dl)


(**** Various Transformations *******************************************)
let flatten_declaration strings kind_and_type =
  List.map (function s -> (s, kind_and_type)) strings
let flatten_sk_declaration names_with_types (kind,restype) =
  List.map
    (function (name,types) -> (name,(types,kind,restype)))
    names_with_types

let unskolemize = List.map (function (n,(ts,k,t)) -> (n,(k,t)))

let make_dec_logical = List.map (function (n,t) -> (n,(Logical,t)))


(***** Checking of declarations *****************************************)

exception Exc_declarations of string

(* check that a type expression t, which occurs in the declaration of *)
(* variable x, is ground, arrow-free, and well-formed.                *)
let check_typeexp_in_decl x t =
  begin
    
    if not (is_ground t)
    then raise (Exc_declarations ("the type expression " ^
				    (string_of_type_expression t) ^
				    " in the declaration of variable " ^
				    x ^ " is not ground."));
    
    if not (arrow_free t)
    then raise (Exc_declarations ("the type expression " ^
				    (string_of_type_expression t) ^
				    " in the declaration of variable " ^
				    x ^ " is not arrow-free."));
    
    try
      check_type_constructors t
    with
	Exc_typecons s ->
	  raise (Exc_declarations
		   ("the type expression in the declaration of " ^
		      "variable " ^ x ^ " " ^ s))
  end


let check_declarations dl =
  begin

    (* check that no variable is declared twice *)
    (try
       ignore(Utils.alldifferent ( List.map (function (x,y) -> x) dl))
     with
	 Utils.Twice x -> raise ( Exc_declarations ("variable \"" ^ x ^
					      "\" declared twice.")));

    (* check that all type expressions are OK *)
    List.iter (function (x,(_,t)) -> check_typeexp_in_decl x t) dl;

  end

(* in order to check a list of skolem declarations we have to       *)
(* - check it as a normal declaration list, ignoring the arguments  *)
(* - check for all the argument types that they are OK.             *)
let check_sk_declarations dl =
  begin

    (* check that no variable is declared twice *)
    (try
       ignore(Utils.alldifferent ( List.map (function (x,y) -> x) dl))
     with
	 Utils.Twice x -> raise ( Exc_declarations ("variable \"" ^ x ^
					      "\" declared twice.")));

    (* check that all type expressions are OK *)
    List.iter (function (x,(args,_,t)) ->
		 begin
		   check_typeexp_in_decl x t;
		   List.iter (check_typeexp_in_decl x) args 
		 end)
      dl

  end
