(**************************************************************************)
(*  ProuveParser: parser frontend for the PROUVÉ language                *)
(*                                                                        *)
(*  Copyright (C) 2004-2009 Ralf Treinen                                  *)
(*                                                                        *)
(*  This library is free software; you can redistribute it and/or         *)
(*  modify it under the terms of the GNU Lesser General Public            *)
(*  License as published by the Free Software Foundation; either          *)
(*  version 2.1 of the License, or (at your option) any later version.    *)
(*                                                                        *)
(*  This library is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *)
(*  Lesser General Public License for more details.                       *)
(*                                                                        *)
(*  You should have received a copy of the GNU Lesser General Public      *)
(*  License along with this library; if not, write to the Free Software   *)
(*  Foundation, Inc., 59 Temple Place, Suite 330,                         *)
(*  Boston, MA  02111-1307  USA                                           *)
(*                                                                        *)
(*  Ralf Treinen <treinen@pps.jussieu.fr>                                 *)
(*  Laboratoire Preuves, Programmes et Systèmes (PPS),                   *)
(*  Université Paris Diderot, Paris, France                              *)
(*  http://www.pps.jussieu.fr/~treinen                                    *)
(*                                                                        *)
(*  This software was mainly written while the author was with the        *)
(*  Laboratoire Spécification et Vérification (LSV),                    *)
(*  ENS de Cachan, Cachan, France                                         *)
(**************************************************************************)

(**********************************)
(*          Expressions           *)
(**********************************)


(* static analysis info for expressions (except identifiers) *)
type var_kind =
  | Var_unknown    (* kind is not yet known *)
  | Var_value      (* can only be an l value *)
  | Var_logical    (* expression which is a logical variable *)
  | Var_mutable    (* expression which is a mutable variable *)
  | Var_of_process (* access to variable of a process *)
  | Var_pid        (* variable denoting a process *)

type info_exp = {
  exp_type : Types.type_expression;
  exp_varkind : var_kind;
  exp_hasvalue : bool
}

val undef_info_exp : info_exp

(* universal type of expressions *)
type expression =
    Ident of info_exp * string
  | Int of info_exp * int
  | Apply of info_exp * string * expression list
  | Infix of info_exp * string * expression * expression
  | Inverse of info_exp * expression
  | Tuple of info_exp * expression list
  | Liste of info_exp * expression list
  | AssocList of info_exp * (expression*expression) list

(* get the info_exp field from an expression *)
val get_info_exp : expression -> info_exp

(* the type of patterns *)
type pattern = expression list

(* conversion to string *)
(* if the first boolean argument is true then the type of each   *)
(* subexpression is indicated in brackets before the expression  *)
val string_of_expression: bool -> expression -> string
val string_of_expression_list: bool -> expression list -> string
val string_of_pattern: bool -> pattern -> string

(* (rename_expr alpha expr) yields a copy of expr where every     *)
(* variable x defined in the association list alpha has been      *) 
(* replaced by alpha(x).                                          *)
(* rename_pattern does the same with patterns.                     *)
val rename_expr: (string * string) list -> expression -> expression
val rename_pattern: (string * string) list -> pattern -> pattern

(* (is_l_expression e) is true iff the expression e is an l-expression *)
(* that is either an identifier, or an identifier applied to some      *)
(* expressions.                                                        *)
val is_l_expression : expression -> bool 

(* collect a sorted list of all integers constants occurring in an *)
(* expression.                                                     *)
val int_constants : expression -> int list

(* collect a sorted list of all variables occurring in an expression.      *)
val vars : expression -> string list
val mutable_vars : expression -> string list
val is_ground : expression -> bool

(* (type_expr sign env sk_env r_env expr) returns the expression expr      *)
(* except that in all sub expresssion of expr (including ecpr itself)      *)
(* the static analysis info has been updated as follows :                  *)
(* - the type field is the (most specific) type of the respective          *)
(*   sub-expression expr                                                   *)
(* - the exp_varkind field has been instantiated.                          *) 
(*  where                                                                  *)
(* - sign is the signature (simple)                                        *)
(* - env is the environment                                                *)
(* - sk_env is the environment for skoelm variables                        *)
(* - r_env is an association list associating to role names its toplevel   *)
(*   declaration list. This is only used when typechecking formulas,       *)
(*   otherwise this is the empty list.                                     *)
(* In an expression resulting from type_expr, the following holds for      *)
(* every subexpression:                                                    *)
(* - the type in an arrow-free type expression without variables.          *)
(* - the type is different from Typeundef                                  *)
(* - the varkind is differnt from Var_unknown                              *)
exception Exc_typeexpression of string
val type_expr:
  Signatures.simple_signature
  -> Declarations.declaration list
  -> Declarations.sk_declaration list
  -> (string * (Declarations.declaration list)) list
  -> expression
  -> expression

val extract_type_expr : expression -> Types.type_expression





