(**************************************************************************)
(*  ProuveParser: parser frontend for the PROUVÉ language                *)
(*                                                                        *)
(*  Copyright (C) 2004-2009 Ralf Treinen                                  *)
(*                                                                        *)
(*  This library is free software; you can redistribute it and/or         *)
(*  modify it under the terms of the GNU Lesser General Public            *)
(*  License as published by the Free Software Foundation; either          *)
(*  version 2.1 of the License, or (at your option) any later version.    *)
(*                                                                        *)
(*  This library is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *)
(*  Lesser General Public License for more details.                       *)
(*                                                                        *)
(*  You should have received a copy of the GNU Lesser General Public      *)
(*  License along with this library; if not, write to the Free Software   *)
(*  Foundation, Inc., 59 Temple Place, Suite 330,                         *)
(*  Boston, MA  02111-1307  USA                                           *)
(*                                                                        *)
(*  Ralf Treinen <treinen@pps.jussieu.fr>                                 *)
(*  Laboratoire Preuves, Programmes et Systèmes (PPS),                   *)
(*  Université Paris Diderot, Paris, France                              *)
(*  http://www.pps.jussieu.fr/~treinen                                    *)
(*                                                                        *)
(*  This software was mainly written while the author was with the        *)
(*  Laboratoire Spécification et Vérification (LSV),                    *)
(*  ENS de Cachan, Cachan, France                                         *)
(**************************************************************************)

open Expressions


(********* Instantiation Analysis ******************************************)

exception Noninst of string

(***** expressions *********************************************************)


let rec is_instantiated iv = function
  | Ident(info,name) ->
      if List.mem name iv || info.exp_varkind = Var_value
      then Ident({info with exp_hasvalue=true}, name) 
      else raise (Noninst name) 
  | Int(info,i) ->
      Int({info with exp_hasvalue=true},i)
  | Apply(info,f,l) ->
      Apply({info with exp_hasvalue=true},
	    f,
	    List.map (is_instantiated iv) l) 
  | Infix(info,op,left,right) ->
      Infix({info with exp_hasvalue=true},
	    op,
	    is_instantiated iv left,
	    is_instantiated iv right)
  | Liste(info,l) ->
      Liste({info with exp_hasvalue=true},
	    List.map (is_instantiated iv) l)
  | Tuple(info,l) ->
      Tuple({info with exp_hasvalue=true},
	    List.map (is_instantiated iv) l)
  | Inverse(info,e) ->
      Inverse({info with exp_hasvalue=true},
	      is_instantiated iv e)
  | AssocList(info,l) ->
      AssocList({info with exp_hasvalue=true},
		List.map
		  (function (key,value) ->
		     (is_instantiated iv key, is_instantiated iv value))
		  l)

(* (mark_instantiated iv expr), where iv is the set of variables that are    *)
(* instantiated, returns a copy of expression expr with the has_value fields *)
(* of all identifiers updated.                                               *)

let rec mark_instantiated iv = function
  | Ident(info,name) ->
      if List.mem name iv || info.exp_varkind = Var_value
      then Ident({info with exp_hasvalue=true}, name) 
      else Ident({info with exp_hasvalue=false}, name)
  | Int(info,i) ->
      Int({info with exp_hasvalue=true},i)
  | Apply(info,f,l) ->
      Apply({info with exp_hasvalue=true},
	    f,
	    List.map (mark_instantiated iv) l) 
  | Infix(info,op,left,right) ->
      Infix({info with exp_hasvalue=true},
	    op,
	    mark_instantiated iv left,
	    mark_instantiated iv right)
  | Liste(info,l) ->
      Liste({info with exp_hasvalue=true},
	    List.map (mark_instantiated iv) l)
  | Tuple(info,l) ->
      Tuple({info with exp_hasvalue=true},
	    List.map (mark_instantiated iv) l)
  | Inverse(info,e) ->
      Inverse({info with exp_hasvalue=true},
	      mark_instantiated iv e)
  | AssocList(info,l) ->
      AssocList({info with exp_hasvalue=true},
		List.map
		  (function (key,value) ->
		     (mark_instantiated iv key, mark_instantiated iv value))
		  l)

(****** patterns ***********************************************************)

let is_matcher hf iv e =
  let rec collect_clauses premisses = function
    | Ident(info,name) -> 
	if info.exp_varkind = Var_logical || info.exp_varkind = Var_mutable
	then [(premisses, name)]
	else []
    | Int _ -> []
    | Infix (_info,op,left,right) ->
	(collect_clauses premisses left) @ (collect_clauses premisses right)
    | Liste(_info,l) | Tuple(_info,l) ->
	List.fold_left (@) [] 
	  (List.map (collect_clauses premisses) l)
    | Inverse(_info,e) -> collect_clauses premisses e
    | AssocList _ -> []
    | Apply(_info,f,l) ->
	if f = "crypt" || f = "sign" || f = "symcrypt"
	then let [algo;key;message]=l
	in
	  collect_clauses
	    (Utils.singlyfy ((vars algo) @ (vars key) @ premisses))
	    message
	else
	  if List.mem f hf
	  then (* f is declared as a hash function *)
	    []
	  else (* f is not declared as a hash function *)
	    List.fold_left (@) [] 
	      (List.map (collect_clauses premisses) l)
  in let rec forward_closure set rules =
      let new_set =
	Utils.singlyfy
	  (List.fold_right
	     (fun (premisses, conc) s -> 
		if (List.for_all (function x -> List.mem x s) premisses)
		then conc::s
		else s)
	     rules
	     set)
      in
	if List.length set = List.length new_set
	then new_set
	else forward_closure new_set rules
  in
  let iv_out = forward_closure iv (collect_clauses [] e)
  and all_vars = vars e
  and all_mutable_vars = mutable_vars e
  in
    (* check whether all mutable variables occuring in pattern are *)
    (* already instantaites before pattern matching.               *)
    (try
       Utils.checksubset all_mutable_vars iv
     with
	 Utils.Missing s -> raise (Noninst s));
    (* check whether the pattern matching instantiates all variables *)
    (* occurring in the pattern                                      *)
    try
      Utils.checksubset all_vars iv_out;
      (e, iv_out)
    with
	Utils.Missing s -> raise (Noninst s)
	    
let rec is_inst_pattern hf iv p =
  let pp = List.map (is_matcher hf iv) p
  in
  let p_inst = List.map (mark_instantiated iv) (List.map fst pp)
  and p_iv = List.map snd pp
  in match p_iv with
    | [] -> [], iv
    | h_iv::r_iv  ->  
	begin 
	  (try List.iter (Utils.checksubset h_iv) r_iv
	   with Utils.Missing v -> raise (Noninst v));
	  (try List.iter (function x -> Utils.checksubset x h_iv) r_iv
	   with Utils.Missing v -> raise (Noninst v))
	end;
	(p_inst, h_iv)

