(**************************************************************************)
(*  ProuveParser: parser frontend for the PROUVÉ language                *)
(*                                                                        *)
(*  Copyright (C) 2004-2009 Ralf Treinen                                  *)
(*                                                                        *)
(*  This library is free software; you can redistribute it and/or         *)
(*  modify it under the terms of the GNU Lesser General Public            *)
(*  License as published by the Free Software Foundation; either          *)
(*  version 2.1 of the License, or (at your option) any later version.    *)
(*                                                                        *)
(*  This library is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *)
(*  Lesser General Public License for more details.                       *)
(*                                                                        *)
(*  You should have received a copy of the GNU Lesser General Public      *)
(*  License along with this library; if not, write to the Free Software   *)
(*  Foundation, Inc., 59 Temple Place, Suite 330,                         *)
(*  Boston, MA  02111-1307  USA                                           *)
(*                                                                        *)
(*  Ralf Treinen <treinen@pps.jussieu.fr>                                 *)
(*  Laboratoire Preuves, Programmes et Systèmes (PPS),                   *)
(*  Université Paris Diderot, Paris, France                              *)
(*  http://www.pps.jussieu.fr/~treinen                                    *)
(*                                                                        *)
(*  This software was mainly written while the author was with the        *)
(*  Laboratoire Spécification et Vérification (LSV),                    *)
(*  ENS de Cachan, Cachan, France                                         *)
(**************************************************************************)

(* formulas *)

open Expressions
open Declarations
open Types

type formula =
    Ftrue
  | Fatom of expression
  | Fsecret of expression
  | Fprocessat of string * string
  | Fscenarioat of string
  | Fand of formula * formula
  | For of formula * formula
  | Fimplies of formula * formula
  | Fnot of formula
  | Fforall of declaration list * formula
  | Fexists of declaration list * formula
  | Funique of declaration list * formula
  | Fbox of formula

(* conversation into string (provided by Liana Bozga) *)
let rec string_of_formula showtypes = function
    Ftrue -> "true"
  | Fatom t -> "("^(Expressions.string_of_expression showtypes t)^")"
  | Fsecret t -> "Secret("^(Expressions.string_of_expression showtypes t)^")"
  | Fprocessat (s1,s2) -> "("^s1^"@"^s2^")"
  | Fscenarioat s -> "(@"^s^")"
  | Fand (f1,f2) ->
      "("^(string_of_formula showtypes f1)
      ^" and "
      ^(string_of_formula showtypes f2)^")"
  | For (f1,f2) ->
      "("^(string_of_formula showtypes f1)
      ^" or "
      ^(string_of_formula showtypes f2)^")"
  | Fimplies (f1,f2) ->
      "("^(string_of_formula showtypes f1)
      ^" => "
      ^(string_of_formula showtypes f2)^")"
  | Fnot f -> "not("^(string_of_formula showtypes f)^")"
  | Fbox f -> "BOX("^(string_of_formula showtypes f)^")"
  | Fforall (st_l, f) ->
      let str_env = Declarations.string_of_parlist st_l in
        "forall "^str_env^(string_of_formula showtypes f)
  | Fexists (st_l, f) ->
      let str_env = Declarations.string_of_parlist st_l in
        "exists "^str_env^(string_of_formula showtypes f)
  | Funique (st_l, f) ->
      let str_env = Declarations.string_of_parlist st_l in
        "unique "^str_env^(string_of_formula showtypes f)


(***** variuous syntactic classes of formulas **************************)

let rec is_quantifier_free = function
    Ftrue | Fatom _ | Fsecret _ | Fprocessat _ | Fscenarioat _ -> true
  | Fnot f -> is_quantifier_free f
  | Fand(f1,f2) | For(f1,f2) | Fimplies(f1,f2) ->
      is_quantifier_free f1 && is_quantifier_free f2
  | _ -> false

(* universal closure of a quantifier-free formula *)
let rec is_pi1 = function
    Fforall (_,f) -> is_pi1 f
  | f -> is_quantifier_free f

(* existential closure of a quantifier-free formula *)
let rec is_sigma1 = function
    Fexists (_,f) | Funique(_,f) -> is_sigma1 f
  | f -> is_quantifier_free f

(* positive boolean combination of sigma1-formulas *)
let rec is_posbool_sigma1 = function
    Fand(f1,f2) | For(f1,f2) -> is_posbool_sigma1 f1 && is_posbool_sigma1 f2
  | f -> is_sigma1 f


(* safety basic formula : universal closure of an implication   *)
(* b -> g where                                                 *)
(* - b is quantifier-free                                       *)
(* - g is a  positive boolean combination of sigma1-formulas    *)
let rec is_safety_basic_formula = function
    Fforall (_,f) -> is_safety_basic_formula f
  | Fimplies (b,g) -> is_quantifier_free b && is_posbool_sigma1 g
  | f -> is_posbool_sigma1 f

(* safety formula: conjunction of safety basic formulas  *)
let rec is_safety_formula = function
    Fand(f1,f2) -> is_safety_formula f1 && is_safety_formula f2
  | f -> is_safety_basic_formula f 



(*********** Type checking *******************************************)

exception Exc_formula of string

let rec typeit sign env r_env r_cp s_cp = function
  | Ftrue as this_formula -> this_formula
  | Fatom e ->
      let e_typed = Expressions.type_expr sign env [] r_env e
      in let e_type = Expressions.extract_type_expr e_typed
      in if e_type = Typeconst "bool"
	then Fatom e_typed
	else raise (Exc_formula
		      ("atomic formula of type " ^ 
			 (string_of_type_expression e_type) ^
			 "instead of type bool."))
  | Fsecret e ->
      let e_typed = Expressions.type_expr sign env [] r_env e
      in Fsecret e_typed
  | Fnot f -> Fnot (typeit sign env r_env r_cp s_cp f)
  | Fbox f -> Fbox (typeit sign env r_env r_cp s_cp f)
  | Fand(f1,f2) -> Fand(
      typeit sign env r_env r_cp s_cp f1,
      typeit sign env r_env r_cp s_cp f2)
  | For(f1,f2) -> For(
      typeit sign env r_env r_cp s_cp f1,
      typeit sign env r_env r_cp s_cp f2)
  | Fimplies(f1,f2) -> Fimplies(
      typeit sign env r_env r_cp s_cp f1,
      typeit sign env r_env r_cp s_cp f2)
  | Funique(dl,f) -> Funique (dl, typeit sign (dl@env) r_env r_cp s_cp f)
  | Fexists(dl,f) -> Fexists (dl, typeit sign (dl@env) r_env r_cp s_cp f)
  | Fforall(dl,f) -> Fforall (dl, typeit sign (dl@env) r_env r_cp s_cp f)
  | Fprocessat(pid,point) as this_formula ->
      begin 
	try 
	  let (kind,pid_type) = List.assoc pid env
	  in if kind <> ProcessKind
	    then raise (Exc_formula
			  ("variable \""^pid^"\" in AT statement " ^
			     "of type \"" ^
			     (string_of_type_expression pid_type)^
			     "\" instead of a role type"))
	    else begin
	      match pid_type with
		  Typeconst(role) ->
		    if List.mem point (List.assoc role r_cp)
		    then
		      this_formula
		    else raise
		      (Exc_formula
			 ("variable \""^pid^"\" in AT statement " ^
			    "at point \""^point^"\" but this is not a " ^
			    "control point of role \""^role^"\""))
		| _ -> failwith
		    "this cannot happen: processkind and no typeconst"
	    end
	with
	    Not_found -> raise (Exc_formula
				  ("variable \""^pid^"\" in AT statement " ^
				     "is not bound"))
      end
  | Fscenarioat(point) as this_formula -> 
      if List.mem point s_cp
      then this_formula
      else raise (Exc_formula
		    ("AT statement for scenario at point \""^point^
		       "\" but this is not a valid checkpoint of the "^
		       "scenario"))
	    
let type_formula sign env r_env r_checkpoints s_checkpoints f = 
  try 
    typeit sign env r_env r_checkpoints s_checkpoints f
  with 
      Expressions.Exc_typeexpression s -> raise (Exc_formula s)
