(**************************************************************************)
(*  ProuveParser: parser frontend for the PROUVÉ language                *)
(*                                                                        *)
(*  Copyright (C) 2004-2009 Ralf Treinen                                  *)
(*                                                                        *)
(*  This library is free software; you can redistribute it and/or         *)
(*  modify it under the terms of the GNU Lesser General Public            *)
(*  License as published by the Free Software Foundation; either          *)
(*  version 2.1 of the License, or (at your option) any later version.    *)
(*                                                                        *)
(*  This library is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *)
(*  Lesser General Public License for more details.                       *)
(*                                                                        *)
(*  You should have received a copy of the GNU Lesser General Public      *)
(*  License along with this library; if not, write to the Free Software   *)
(*  Foundation, Inc., 59 Temple Place, Suite 330,                         *)
(*  Boston, MA  02111-1307  USA                                           *)
(*                                                                        *)
(*  Ralf Treinen <treinen@pps.jussieu.fr>                                 *)
(*  Laboratoire Preuves, Programmes et Systèmes (PPS),                   *)
(*  Université Paris Diderot, Paris, France                              *)
(*  http://www.pps.jussieu.fr/~treinen                                    *)
(*                                                                        *)
(*  This software was mainly written while the author was with the        *)
(*  Laboratoire Spécification et Vérification (LSV),                    *)
(*  ENS de Cachan, Cachan, France                                         *)
(**************************************************************************)

(************** abstract sytntax of roles **********************************)

open Expressions
open Types
open Declarations

type rawinstruction =
  | IRawLabeled of string * rawinstruction
  | IRawBlock of rawblock
  | IRawSend of expression
  | IRawRecv of pattern
  | IRawFail
  | IRawAssign of string * expression
  | IRawNew of string
  | IRawMakeKey of expression * string
  | IRawMatch of expression * pattern
  | IRawConditional of expression * rawinstruction list * rawinstruction list
  | IRawChoice of rawinstruction list list
  | IRawCase of expression * (pattern * rawinstruction list) list
      * rawinstruction list
and rawblock =
    declaration list * rawinstruction list

type instruction =
  | IBlock of string*block
  | ISend of string*expression
  | IRecv of string*pattern
  | IFail of string
  | IAssign of string * string * expression
  | INew of string * string
  | IMakeKey of string  * expression * string
  | IMatch of string * expression * pattern
  | IConditional of string * expression * instruction list * instruction list
  | IChoice of string * (instruction list) list
  | ICase of string *
      expression * (pattern * instruction list) list * instruction list
and block =
    declaration list * instruction list

type role = 
    Role of string * declaration list * block

(*********** processing ***********************************************)

let rec cook_block label (dlist,instlist) =
  (dlist,cook_ilist label instlist)
and cook_ilist label =
   Utils.mapindex (function n -> cook_inst (label^"_"^(string_of_int n)))
and cook_inst label = function
  | IRawLabeled(newlabel,inst) -> cook_inst newlabel inst
  | IRawBlock block -> IBlock(label,cook_block label block)
  | IRawSend exp -> ISend(label,exp)
  | IRawRecv pat -> IRecv(label,pat)
  | IRawFail -> IFail(label)
  | IRawAssign(var,exp) -> IAssign(label,var,exp)
  | IRawNew(var) -> INew(label,var)
  | IRawMakeKey(var,exp) -> IMakeKey(label,var,exp)
  | IRawMatch(exp,pat) -> IMatch(label,exp,pat)
  | IRawConditional(cond,thenpart,elsepart) ->
      IConditional(
	label,
	cond,
	(cook_ilist (label^"_thn") thenpart),
	(cook_ilist (label^"_els") elsepart))
  | IRawChoice choicelist ->
      IChoice(
	label,
	(Utils.mapindex 
	   (fun n ilist -> 
	      (cook_ilist (label^"_"^(string_of_int n)) ilist))
	   choicelist))
  | IRawCase(exp,caslist,defaultlist) ->
      ICase(
	label,
	exp,
	(Utils.mapindex 
	   (fun n (pat,ilist) -> 
	      (pat,
	       cook_ilist (label^"_"^(string_of_int n)) ilist))
	   caslist),
	cook_ilist (label^"_dft") defaultlist)
	  
    
let process_instblock labelprefix block = cook_block labelprefix block

(*********** printing *************************************************)

let rec print_block showtype oc prefix (decllist,instlist) =
  output_string oc "declare\n";
  List.iter
    (function decl ->
       output_string oc (prefix^"  ");
       output_string oc (string_of_declaration decl);
       output_string oc ";\n")
    decllist;
  output_string oc prefix;
  output_string oc "begin\n";
  print_instruction_list showtype oc (prefix^"  ") instlist;
  output_string oc prefix;
  output_string oc "end";
and print_instruction_list showtype oc prefix = List.iter
  (function inst -> 
     output_string oc prefix;
     print_instruction showtype oc prefix inst;
     output_string oc ";\n")
and print_instruction showtype oc prefix = function
  | ISend (label,exp) ->
      output_string oc (label^": ");
      output_string oc "send(";
      output_string oc (string_of_expression showtype exp);
      output_char oc ')'
  | IRecv (label,pat) ->
      output_string oc (label^": ");
      output_string oc "recv(";
      output_string oc (string_of_pattern showtype pat);
      output_char oc ')'
  | IFail label ->
      output_string oc (label^": ");
      output_string oc "fail"
  | IAssign(label,var,exp) ->
      output_string oc (label^": ");
      output_string oc var;
      output_string oc " := ";
      output_string oc (string_of_expression showtype exp)
  | INew(label,var) ->
      output_string oc (label^": ");
      output_string oc "new(";
      output_string oc var;
      output_char oc ')'
  | IMakeKey(label,expr,var) ->
      output_string oc (label^": ");
      output_string oc "makekey(";
      output_string oc (string_of_expression showtype expr);
      output_string oc ", ";
      output_string oc var;
      output_char oc ')'
  | IConditional(label,cond,thenpart,elsepart) ->
      output_string oc (label^": ");
      output_string oc "if ";
      output_string oc (string_of_expression showtype cond);
      output_string oc " then\n";
      print_instruction_list showtype oc (prefix^"  ") thenpart;
      output_string oc prefix;
      output_string oc "else\n";
      print_instruction_list showtype oc (prefix^"  ") elsepart;
      output_string oc prefix;
      output_string oc "fi"
  | IMatch(label,exp,pat) ->
      output_string oc (label^": ");
      output_string oc "match ";
      output_string oc (string_of_expression showtype exp);
      output_string oc " with ";
      output_string oc (string_of_pattern showtype pat);
      output_string oc " end"
  | IBlock (label,bl) ->
      output_string oc (label^": ");
      print_block showtype oc prefix bl
  | IChoice(label,choicelist) ->
      output_string oc (label^": ");
      output_string oc "choice\n ";
      List.iter
	(function instl ->
	   output_string oc (prefix^"  |\n");
	   print_instruction_list showtype oc (prefix^"      ") instl;
	   output_char oc '\n')
	choicelist;
      output_string oc prefix;
      output_string oc "end"
  | ICase(label,exp,caselist,elsepart) ->
      output_string oc (label^": ");
      output_string oc "case ";
      output_string oc (string_of_expression showtype exp);
      output_string oc " of ";
      List.iter
	(function (pat,instl) ->
	   output_string oc (prefix^"  ");
	   output_string oc (string_of_pattern showtype pat);
	   output_string oc " ->\n";
	   print_instruction_list showtype oc (prefix^"      ") instl;
	   output_char oc '\n')
	caselist;
      output_string oc prefix;
      output_string oc "else\n";
      print_instruction_list showtype oc (prefix^"  ") elsepart;
      output_string oc prefix;
      output_string oc "fi"



let print_role showtype oc (Role (rolename, decllist, bl)) =
  output_string oc "role ";
  output_string oc rolename;
  output_string oc " (";
  output_string oc (string_of_parlist decllist);
  output_string oc ")\n";
  print_block showtype oc "" bl;
  output_string oc "\n\n"
    
;; 

(************ renaming of variables ********************************)

let rec rename_ilist alpha = List.map (rename_inst alpha)
and rename_inst alpha = function
  | IBlock(label,(dl,il)) ->
      let nl = List.map fst dl
	(* newalpha is alpha without the variables declared in dl *)
      in let newalpha =
	  List.filter
	    (function (x,_) -> not (List.mem x nl))
	    alpha
      in
	IBlock(label,(dl, rename_ilist newalpha il))
  | INew (label,x) as this ->
      (try
	 INew (label,List.assoc x alpha)
       with
	   Not_found -> this)
  | IMakeKey (label,e,x) ->
      IMakeKey(label,
	       rename_expr alpha e,
	       (try List.assoc x alpha with Not_found -> x))
  | IFail label -> IFail label
  | ISend (label,e) -> ISend (label,rename_expr alpha e)
  | IRecv (label,p) -> IRecv (label,rename_pattern alpha p)
  | IAssign (label,x,e) ->
      IAssign (label,
	       (try List.assoc x alpha with Not_found -> x),
	       (rename_expr alpha e))
  | IMatch (label,e,p) -> 
      IMatch (label,
	      (rename_expr alpha e),
	      (rename_pattern alpha p))
  | IConditional (label,cond, thenpart, elsepart) ->
      IConditional (label,
		    (rename_expr alpha cond),
		    (rename_ilist alpha thenpart),
		    (rename_ilist alpha elsepart))
  | IChoice (label,choicelist) ->
      IChoice (label,
	       (List.map (rename_ilist alpha) choicelist))
  | ICase (label,e,caselist,elsepart) ->
      ICase (label,
	     (rename_expr alpha e),
	     (List.map
		(function (p,il) ->
		   (rename_pattern alpha p),
		   (rename_ilist alpha il))
		caselist),
	     (rename_ilist alpha elsepart))
	
(*******************************************************************)

let rec bound_vars_inst = function
    IBlock(_,(dl,il)) -> (List.map fst dl) @ (bound_vars_ilist il)
  | INew _ | ISend _ | IRecv _ | IAssign _ | IFail _ | IMatch _ 
  | IMakeKey _ -> []
  | IConditional(_,_,thenpart,elsepart) ->
      (bound_vars_ilist thenpart) @ (bound_vars_ilist elsepart)
  | IChoice(_,choicelist) ->
      (List.flatten (List.map bound_vars_ilist choicelist))
  | ICase(_,_,caselist,elsepart) ->
      (List.flatten 
	 (List.map (function (p,il) -> bound_vars_ilist il) caselist)) @
	(bound_vars_ilist elsepart)
and bound_vars_ilist il = 
  List.flatten (List.map bound_vars_inst il)
	 

let rec get_form_parlist name = function
    [] -> raise Not_found
  | Role(n,pars,_):: l ->
      if n = name
      then List.map fst pars
      else get_form_parlist name l

let gather_parameter_declarations =
  List.map ( function Role(name,decl_list,_) -> (name,decl_list) )

let gather_toplevel_declarations =
  List.map
    ( function Role(name,decl_list,(block_decl,_)) ->
	(name, decl_list@block_decl) )


(************ collect checkpoints *****************************************)

let rec checkpoints_il il =
  List.fold_right (fun x y -> (checkpoints x)@y) il [] 
and checkpoints = function
  | ISend(l,_) | IRecv(l,_) | IFail(l) | IAssign(l,_,_) | INew(l,_)
  | IMakeKey(l,_,_) | IMatch(l,_,_) -> [l]
  | IBlock(l,(_,il)) -> l :: (checkpoints_il il)
  | IConditional(l,_,thenpart,elsepart) ->
      l::(checkpoints_il thenpart)@(checkpoints_il elsepart)
  | IChoice(l,clauselist) -> 
      l::(List.fold_right (fun x y -> (checkpoints_il x)@y) clauselist [])
  | ICase(l,_,clauselist,elselist) ->
      l::(List.fold_right (fun (_,x) y -> (checkpoints_il x)@y) clauselist [])
      @(checkpoints_il elselist)
    
let gather_checkpoints =
  List.map (function Role(name,_,(_,il)) ->
	      (name, ["start"; "end"] @ (checkpoints_il il)))

(*************  collect integer constants  ********************************)

let ic_e = Expressions.int_constants
let ic_el l = List.flatten(List.map ic_e l)
let rec ic_inst = function
    IBlock(_,b) -> ic_block b
  | ISend(_,e) -> ic_e e
  | IRecv(_,l) -> ic_el l
  | IFail _ -> []
  | IAssign(_,_,e) -> ic_e e
  | INew _ -> []
  | IMakeKey(_,e,_) -> ic_e e 
  | IMatch(_,e,l) -> (ic_e e) @ (ic_el l)
  | IConditional(_,cond,il_then,il_else) ->
      (ic_e cond) @ (ic_list il_then) @ (ic_list il_else)
  | IChoice(_,choicelist) ->
      (List.flatten (List.map ic_list choicelist))
  | ICase(_,exp,caselist,default) ->
      (ic_e exp) @
      (List.flatten
	 (List.map
	    (function (el,il) -> (ic_el el) @ (ic_list il))
	    caselist)) @
      (ic_list default)
and ic_block b = ic_list (snd b)
and ic_list l =  List.flatten (List.map ic_inst l)
    
let int_constants (Role(name,decl,block)) =
  Utils.singlyfy (ic_block block)

(***************** Typing *************************************************)

exception Exc_roles of string

let rec type_block sign env (decl_list, inst_list) =
  (try
     (check_declarations decl_list; decl_list,
      type_instlist sign (decl_list@env) inst_list)
   with
       Exc_declarations s -> raise
	 (Exc_roles 
	    ("Error in declaration part of a block: " ^ s)))

and type_instlist sign env =
  List.map (type_inst sign env)

and type_inst sign env = function

  | IBlock(label,b) -> IBlock(label, type_block sign env b)

  | ISend(label,e) ->
      (try
	 ISend(label, type_expr sign env [] [] e)
       with
	   Exc_typeexpression s -> raise 
	     (Exc_roles
		("type error in the argument of a send instruction: " ^ s)))

  | IRecv(label,pl) ->
      (try
	 IRecv(label, List.map (type_expr sign env [] []) pl)
       with
	   Exc_typeexpression s -> raise 
	     (Exc_roles
		("type error in a pattern of of a recv instruction: "
		 ^ s)))
	   
  | IFail _label as inst -> inst

  | IConditional(label,cond,thenlist,elselist) ->
      (try
	 let cond_typed = type_expr sign env [] [] cond
	 in let cond_type = extract_type_expr cond_typed
	 in if cond_type = Typeconst "bool"
	   then
	     IConditional(label,cond_typed,
			  type_instlist sign env thenlist,
			  type_instlist sign env elselist)
	   else raise (Exc_roles
			 ("conditional with guard \"" ^
			  (string_of_expression false cond) ^
			  "\" which has type " ^
			  (string_of_type_expression cond_type) ^
			  " instead of bool"))
       with
	   Exc_typeexpression s -> raise
	     (Exc_roles
		("condition \"" ^ (string_of_expression false cond) ^
		 "\" can not be typed because: " ^ s)))

  | IAssign(label, varname, exp) ->
      (try
	 let var_type = snd (List.assoc varname env)
	 and exp_typed = type_expr sign env [] [] exp 
	 in let exp_type = extract_type_expr exp
	 in if smaller_type exp_type var_type
	   then	IAssign(label, varname, exp_typed)
	   else raise (Exc_roles
			 ("assignment of \"" ^
			    (string_of_expression false exp) ^
			  "\" of type " ^
			  (string_of_type_expression exp_type) ^
			  " to variable \"" ^ varname ^
			  "\" of type " ^
			  (string_of_type_expression var_type)))
       with
	   Not_found -> raise (Exc_roles
				 ("assignment to non declared variable \"" ^
				  varname ^ "\""))
	 | Exc_typeexpression s -> raise (Exc_roles
					   ("expression \"" ^ 
					    (string_of_expression false exp) ^
					    "\" cannot be typed because "^ s)))

  | INew(label,varname) as inst ->
      (try
	 let var_type = snd (List.assoc varname env)
	 in if (var_type = Typeconst "nonce"
	       || var_type = Typeconst "int"
	       || var_type = Typeconst "message" )
	   then inst
	   else
	     raise (Exc_roles
		      ("\"new\" instruction on variable \"" ^ varname ^
			 "\" which is of type " ^ 
			 (string_of_type_expression var_type) ^ 
			 "but must be one of \"nonce\" or \"int\""))
       with
	   Not_found -> raise (Exc_roles
				 ("new on a non declared variable \"" ^
				    varname ^ "\"")))

  | IMakeKey(label,alg_arg,pkey_arg) ->
      let alg_arg_typed = 
	try type_expr sign env [] [] alg_arg
	with
	    Exc_typeexpression s -> 
	      raise (Exc_roles
		       ("cannot type first argument \"" ^
			  (string_of_expression false alg_arg) ^
			  "\" of makekey because " ^ s))
      in
      let alg_arg_type = extract_type_expr alg_arg_typed
      and pkey_arg_type = 
	try snd (List.assoc pkey_arg env)
	with
	    Not_found -> 
	      raise (Exc_roles
		       ("the second argument \"" ^ pkey_arg ^
			  "\" of makekey is not declared"))
      in
	(match alg_arg_type with
	     Typeconst "algo" ->
	       if pkey_arg_type = Typeconst "privkey"
	       then
		 IMakeKey(label,alg_arg_typed,pkey_arg)
	       else raise
		 (Exc_roles
		    ("second argument \"" ^
		       pkey_arg ^
		       "\" of makekey has type " ^
		       (string_of_type_expression pkey_arg_type) ^
		       " instead of privkey"))
	   | Typeconst "symalgo" ->
	       if pkey_arg_type = Typeconst "symkey"
	       then
		 IMakeKey(label,alg_arg_typed,pkey_arg)
	       else raise
		 (Exc_roles
		    ("second argument \"" ^
		       pkey_arg ^
		       "\" of makekey has type " ^
		       (string_of_type_expression pkey_arg_type) ^
		       " instead of symkey"))
	   | _ -> raise
	       (Exc_roles
		  ("first argument \"" ^
		     (string_of_expression false alg_arg) ^
		     "\" of makekey has type " ^
		     (string_of_type_expression alg_arg_type) ^
		     " instead of algo or symalgo")))

  | IChoice(label,choicelist) ->
      IChoice(label, List.map (type_instlist sign env) choicelist) 

  | IMatch(label,expr,patterns) ->
      (try
	 let expr_typed = type_expr sign env [] [] expr
	 in let expr_type = extract_type_expr expr_typed
	 in IMatch(label, expr_typed,
		   type_patterns sign env expr_type patterns)
       with
	   Exc_typeexpression s ->
	     raise (Exc_roles
		      ("expression \"" ^ 
			 (string_of_expression false expr) ^
			 "\" cannot be typed because "^ s)))

  | ICase(label,expr,caselist,default) ->
      (try
	 let expr_typed = type_expr sign env [] [] expr
	 in let expr_type = extract_type_expr expr_typed
	 in 
	   ICase(label,
		 expr_typed,
		 List.map
		   (function (patterns,insts) ->
		      (type_patterns sign env expr_type patterns,
		       type_instlist sign env insts))
		   caselist,
		 type_instlist sign env default)
       with
	   Exc_typeexpression s ->
	     raise (Exc_roles
		      ("expression \"" ^ 
			 (string_of_expression false expr) ^
			 "\" cannot be typed because "^ s)))

(* (type_patterns sign env exp_type patterns) verifies that, for signature *)
(* sign and variable declarations env,                                     *)
(* - every pattern in patterns is typable                                  *)
(* - the type of any pattern p has an infimum with the type of exp_type    *)
(* if this is the case then the function returns the same patterns with    *)
(* all type and kind annotations.                                          *)
  and type_patterns sign env exp_type patterns =
  List.map
    (function p -> 
       try
	 let p_typed = type_expr sign env [] [] p
	 in let p_type = extract_type_expr p_typed
	 in 
	   if has_inf_type exp_type p_type
	   then p_typed
	   else raise (Exc_roles
			 ("pattern \"" ^
			    (string_of_expression false p) ^
			    "\" has type " ^
			    (string_of_type_expression p_type) ^
			    " which does not compare with type \"" ^
			    (string_of_type_expression exp_type)^"\""))
       with
	   Exc_typeexpression s ->
	     raise (Exc_roles
		      ( "pattern \"" ^ 
			  (string_of_expression false p) ^
			  "\" cannot be typed because " ^ s)))
    patterns


let check_rolename name =
  if List.mem name Types.predefined_type_constants
  then raise
    (Exc_roles ("the name of this role clashes with the name" ^
		  " of a pre-defined type constant."))

(* check a list of roles *)
let type_roles sign rolelist=
  (* check that all checkpoints in roles are unique *)
  List.iter
    (function (rolename,checkpoint_list) ->
       try
	 Utils.alldifferent checkpoint_list
       with
	   Utils.Twice cp ->
	     raise
	       (Exc_roles
		  ("the checkpoint \"" ^cp^ "\" occurs multiple times " ^
		   "in role \"" ^ rolename ^"\"")))
    (gather_checkpoints rolelist);
  
  (* now go over the list of role definitiosn and type each of them. *)
  List.map
    (function Role(name,decls,block) ->
       try
	 check_rolename name;
	 check_declarations decls;
	 Role(name,decls,type_block sign decls block)
       with
	   Exc_declarations s -> raise
	     (Exc_roles 
		("error in the parameter list of role \"" ^ name ^ "\":\n"
		 ^ s))
	 | Exc_roles s -> raise
	     (Exc_roles
		("error in the body of role \"" ^ name ^ "\":\n" ^ s)))
    rolelist


	

  
    

    
