(**************************************************************************)
(*  ProuveParser: parser frontend for the PROUVÉ language                *)
(*                                                                        *)
(*  Copyright (C) 2004-2009 Ralf Treinen                                  *)
(*                                                                        *)
(*  This library is free software; you can redistribute it and/or         *)
(*  modify it under the terms of the GNU Lesser General Public            *)
(*  License as published by the Free Software Foundation; either          *)
(*  version 2.1 of the License, or (at your option) any later version.    *)
(*                                                                        *)
(*  This library is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *)
(*  Lesser General Public License for more details.                       *)
(*                                                                        *)
(*  You should have received a copy of the GNU Lesser General Public      *)
(*  License along with this library; if not, write to the Free Software   *)
(*  Foundation, Inc., 59 Temple Place, Suite 330,                         *)
(*  Boston, MA  02111-1307  USA                                           *)
(*                                                                        *)
(*  Ralf Treinen <treinen@pps.jussieu.fr>                                 *)
(*  Laboratoire Preuves, Programmes et Systèmes (PPS),                   *)
(*  Université Paris Diderot, Paris, France                              *)
(*  http://www.pps.jussieu.fr/~treinen                                    *)
(*                                                                        *)
(*  This software was mainly written while the author was with the        *)
(*  Laboratoire Spécification et Vérification (LSV),                    *)
(*  ENS de Cachan, Cachan, France                                         *)
(**************************************************************************)

(***** Instantiation analysis *********************************************)

open Expressions
open Expressions_analysis
open Declarations
open Roles

(* (not_always_fails inst_list) returns true iff there exists an *)
(* execution path of the instruction list inst_list which does   *)
(* not contain the instruction "fail".                           *)

let rec not_always_fails = function
  | [] -> true
  | h::r -> match h with
      | IFail _ -> false
      | ISend _ | IRecv _ | IAssign _ | INew _ | IMatch _ | IMakeKey _ ->
	  not_always_fails r
      | IConditional(_label,_cond,thenpart,elsepart) ->
	  not_always_fails thenpart || not_always_fails elsepart
      | IChoice(_label,ill) ->
	  List.exists not_always_fails ill
      | ICase(_label, _expr, cl, defaultpart) ->
	  (List.exists (function (_pat,il) -> not_always_fails il) cl)
	  || not_always_fails defaultpart
      | IBlock(_label,(_decls,il)) ->
	  not_always_fails il

(* Auxiliary functions for instantiation analysis:                  *)
(* ia_block, ia_inst, ia_instalist, ia_role.                        *)
(* In these functions, iv is always the list of variables which are *)
(* already instantiated, and env is the current environment.        *)
(* hf is the list of hash functions.                                *)
(* The result of these functions is always a pair, consisting of:   *)
(* - an entity of the abtstract syntax of the same type as the last *)
(*   argument of the function                                       *)
(* - a list of variables whcih are instantiated *after* successful  *)
(*   execution of that piece of a role.                             *)

let rec ia_block hf env iv (decl_list, inst_list) =
  (* analysis of a block: given annviroment env and a set of instantiated   *)
  (* variables iv, we have to                                               *)
  (* - analyse the body of the block with the environmen extended by the    *)
  (*   local declarations of the block, and a set of instantaited variables *)
  (*   which is iv minus the varaibles locally declated in this block.      *)
  (* - return a set of instantiated varaibles which is the one obtained     *)
  (*   from the analysis of the body of the block, minus those which are    *)
  (*   are locally decalred in the block, plus those which were already     *)
  (*   instantiated at the beginning of the block.                          *)
  (* Note that the property of a variable being instantiate is monotonous   *)
  (* along execution of a role, but we have to keep track of local          *)
  (* declarations and visibility rules.                                     *)
  let local_variables = Utils.singlyfy (List.map fst decl_list)
  in let inst_list_ia, ivr = 
     ia_instlist
     hf
     (decl_list@env)
     (Utils.listminus (Utils.singlyfy iv) local_variables)
     inst_list
  in
    ((decl_list, inst_list_ia),
     (Utils.singlyfy
	(iv@(Utils.listminus (Utils.singlyfy ivr) local_variables))))
and ia_inst hf env iv = function
  | IBlock(label, block) -> 
      let (block_ia, ivr) = ia_block hf env iv block
      in  IBlock(label, block_ia), ivr
  | ISend(label, exp) ->
      let exp_ia = 
	(try
	   is_instantiated iv exp
	 with
	     Noninst v ->
	       raise (Exc_roles 
			("the variable \"" ^ v ^ "\" is not instantiated in " ^
			   " instruction send(" ^
			   (string_of_expression false exp) ^ ")")))
      in ISend(label, exp_ia), iv
  | INew(label, varname) as this_inst ->
      if get_kind_decl varname env = Logical && List.mem varname iv
      then raise (Exc_roles ("instruction \"new\" on variable " ^ varname ^
			       " which is logical and already instantiated"))
      else
	this_inst, if List.mem varname iv then iv else varname::iv
  | IMakeKey(label, expr, varname) as this_inst ->
      if get_kind_decl varname env = Logical && List.mem varname iv
      then raise (Exc_roles ("instruction \"makekey\" on variable " ^ varname ^
			       " which is logical and already instantiated"))
      else
	this_inst, if List.mem varname iv then iv else varname::iv
  | IAssign(label, varname, exp) ->
      if get_kind_decl varname env = Logical && List.mem varname iv
      then
	raise (Exc_roles ("asignment on variable " ^ varname ^
			    " which is logical and already instantiated"))
      else
	let exp_ia =
	  (try is_instantiated iv exp
	   with
	       Noninst v ->
		 raise (Exc_roles 
			  ("the variable \"" ^ v ^
			     "\" is not instantiated in " ^
			     " assignment of " ^
			     (string_of_expression false exp) ^
			     " to variable " ^ varname)))
	in
	  (IAssign(label, varname, exp_ia),
	   if List.mem varname iv then iv else varname::iv)
  | IFail _label as this_inst -> this_inst, iv
  | IConditional(label, cond, thenpart, elsepart) ->
      let cond_ia =
	(try is_instantiated iv cond
	 with
	     Noninst v ->
	       raise (Exc_roles 
			("the variable \"" ^ v ^ "\" is not instantiated in " ^
			   " a conditional instruction  with guard " ^
			   (string_of_expression false cond))))
      in
      let thenpart_ia, iv_then = ia_instlist hf env iv thenpart
      and elsepart_ia, iv_else = ia_instlist hf env iv elsepart
      in
	if not_always_fails thenpart
	then if not_always_fails elsepart
	then begin
	  (* we have to check whether after both branches the same *)
	  (* variables are instantiated.                           *)
	  (try Utils.checksubset iv_then iv_else
	   with Utils.Missing v -> 
	     raise (Exc_roles
		      ("the variable " ^ v ^
			 " is instantiated in a then-branch"
		       ^ " but not in the corresponding else-branch of a "
		       ^ "conditional instruction")));
	  (try Utils.checksubset iv_else iv_then
	   with Utils.Missing v -> 
	     raise (Exc_roles
		      ("the variable " ^ v ^
			 " is instantiated in a else-branch"
		       ^ " but not in the corresponding then-branch of a "
		       ^ "conditional instruction")));
	  (IConditional(label, cond_ia, thenpart_ia, elsepart_ia), iv_then)
	end
	else  (* elsepart always fails *)
	  (IConditional(label, cond_ia, thenpart_ia, elsepart_ia), iv_then)
	else (* thenpart always fails *)
	  (IConditional(label, cond_ia, thenpart_ia, elsepart_ia), iv_else)
  | IChoice(label,l) ->
      let l_anno = List.map (ia_instlist hf env iv) l
      in let l_ia = List.map fst l_anno
	(* we construct the list of instantiated variables only *)
	(* for those branches that do not always fail.          *)
      and l_iv_indexed =
	List.map
	  (function (index,(_,iv)) -> (index,iv))
	  (Utils.filter_index
	     (function (x,y) -> not_always_fails x)
	     l_anno)
      in
	begin
	  match l_iv_indexed with
	    | [] ->
		(IChoice(label, l_ia), iv)
	    | (h_index,h_iv)::r ->
		List.iter
		  (function (index,iv) ->
		     try
		       Utils.checksubset h_iv iv
		     with Utils.Missing v -> 
		       raise
			 (Exc_roles
			    ("the variable " ^ v ^ " is instantiated in " ^
			       "branch " ^ (string_of_int h_index) ^
			       " of a choice instruction but is " ^
			       "not instantiated in branch " ^
			       (string_of_int index) ^
			       " of the same choice instruction.")))
		  r;
		List.iter
		  (function (index,iv) ->
		     try
		       Utils.checksubset iv h_iv
		     with Utils.Missing v -> 
		       raise
			 (Exc_roles
			    ("the variable " ^ v ^ " is instantiated in " ^
			       "branch " ^ (string_of_int index) ^
			       " of a choice instruction but is " ^
			       "not instantiated in branch " ^
			       (string_of_int h_index) ^
			       " of the same choice instruction.")))
		  r;
		(IChoice(label, l_ia), h_iv)
	end
  | IRecv(label,p) -> 
      let p_ia,p_iv =
	try
	  is_inst_pattern hf iv p
	with
	    Noninst v ->
	      raise (Exc_roles 
		       ("the variable \"" ^ v ^ "\" is not instantiated " ^
			  "in a recv instruction with pattern "^
			  (string_of_pattern false p)))
      in
	(IRecv(label,p_ia), p_iv)
  | IMatch(label,exp,p) -> 
      let p_ia,p_iv =
	try
	  is_inst_pattern hf iv p
	with
	    Noninst v ->
	      raise (Exc_roles 
		       ("the variable \"" ^ v ^ "\" is not instantiated " ^
			  "in a match instruction with pattern "^
			  (string_of_pattern false p)))
      and exp_ia =
	try
	  is_instantiated iv exp
	with
	    Noninst v ->
	      raise (Exc_roles 
		       ("the variable \"" ^ v ^
			  "\" is not instantiated in " ^
			  " matching with expression " ^
			  (string_of_expression false exp)))
      in
	(IMatch(label,exp_ia,p_ia), p_iv)
  | ICase(label,exp,cl,default) ->
      let exp_ia =
	try
	  is_instantiated iv exp
	with
	    Noninst v ->
	      raise (Exc_roles 
		       ("the variable \"" ^ v ^
			  "\" is not instantiated in " ^
			  " case statement with expression " ^
			  (string_of_expression false exp)))
      and default_ia, default_iv = ia_instlist hf env iv default
      and cll =
	List.map
	  (function (pat,il) ->
	     let (pat_ia,pat_iv) = is_inst_pattern hf iv pat
	     in let (il_ia, il_iv) = ia_instlist hf env pat_iv il
	     in ((pat_ia, il_ia), il_iv))
	  cl
      in
      let cl_ia = List.map fst cll
	(* we construct the list of instantiated variables only *)
	(* for those branches that do not always fail.          *)
      and cl_iv =
	List.map snd
	  (List.filter
	     (function ((_pat,il),_iv) -> not_always_fails il)
	     cll)
      in
	begin 
	  (try List.iter (Utils.checksubset default_iv) cl_iv
	   with Utils.Missing v -> 
	     raise
	       (Exc_roles
		  ("the variable " ^ v ^ "is instantiated in the " ^
		     "default branch of a case instruction but is " ^
		     "not instantiated in some regular branch")));
	  (try List.iter (function x -> Utils.checksubset x default_iv) cl_iv
	   with Utils.Missing v -> 
	     raise
	       (Exc_roles
		  ("the variable " ^ v ^ "is not instantiated in " ^
		     "the default branch of a match instruction but " ^
		     "is instantiated in some regular branch")));
	end;
	(ICase(label, exp_ia, cl_ia, default_ia), default_iv)

and ia_instlist hf env iv = function
  | [] -> [],iv
  | h::r -> 
      let (h_ia,ivr) = ia_inst hf env iv h
      in let (r_ia,ivrr) = ia_instlist hf env ivr r
      in (h_ia::r_ia), ivrr

let ia_role hf (Role(rolename, decllist, block)) =
  try
    let block_ia, _ =
      ia_block
	hf
	decllist
	(* when the main block of a role is analysed then the set of *)
	(* variables instantiated at the the beginning of the block  *)
	(* is the set of *all* parameters of the role.               *)
	(* this is an OVERAPPROXMATION                               *)
	(List.map fst
	   (*
	     (List.filter
	     (function (_name, (kind,_type)) -> kind=Logical)
	     decllist))
	   *)
	   decllist)
	block
    in
      Role(rolename, decllist, block_ia)
  with
      Exc_roles s -> 
	raise (Exc_roles
		 ("instantiation error in role " ^ rolename ^
		    ": " ^s))

(* the main function for instantiation analysis of roles.                *)
let instantiate_roles hf = List.map (ia_role hf)
