(**************************************************************************)
(*  ProuveParser: parser frontend for the PROUVÉ language                *)
(*                                                                        *)
(*  Copyright (C) 2004-2009 Ralf Treinen                                  *)
(*                                                                        *)
(*  This library is free software; you can redistribute it and/or         *)
(*  modify it under the terms of the GNU Lesser General Public            *)
(*  License as published by the Free Software Foundation; either          *)
(*  version 2.1 of the License, or (at your option) any later version.    *)
(*                                                                        *)
(*  This library is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *)
(*  Lesser General Public License for more details.                       *)
(*                                                                        *)
(*  You should have received a copy of the GNU Lesser General Public      *)
(*  License along with this library; if not, write to the Free Software   *)
(*  Foundation, Inc., 59 Temple Place, Suite 330,                         *)
(*  Boston, MA  02111-1307  USA                                           *)
(*                                                                        *)
(*  Ralf Treinen <treinen@pps.jussieu.fr>                                 *)
(*  Laboratoire Preuves, Programmes et Systèmes (PPS),                   *)
(*  Université Paris Diderot, Paris, France                              *)
(*  http://www.pps.jussieu.fr/~treinen                                    *)
(*                                                                        *)
(*  This software was mainly written while the author was with the        *)
(*  Laboratoire Spécification et Vérification (LSV),                    *)
(*  ENS de Cachan, Cachan, France                                         *)
(**************************************************************************)

open Scenarios
open Types
open Expressions
open Declarations
open Intervals

let rec type_scenario sign role_decl gl_var_decl env = function
    SCall(label,"makekey",arguments) ->
      (match arguments with
	   [alg_arg; pkey_arg] ->
	     let alg_arg_typed = 
	       try type_expr sign env gl_var_decl [] alg_arg
	       with
		   Exc_typeexpression s -> 
		     raise (Exc_scenario
			      ("cannot type first argument \"" ^
				 (string_of_expression false alg_arg) ^
				 "\" of makekey because " ^ s))
	     in
	     let alg_arg_type = extract_type_expr alg_arg_typed
	     and pkey_arg_typed =
	       try type_expr sign env gl_var_decl [] pkey_arg
	       with
		   Exc_typeexpression s -> 
		     raise (Exc_scenario
			      ("cannot type second argument \"" ^
				 (string_of_expression false pkey_arg) ^
				 "\" of makekey because " ^ s))
	     in
	     let pkey_arg_type = extract_type_expr pkey_arg_typed
	     in
	       (match alg_arg_type with
		    Typeconst "algo" ->
		      if pkey_arg_type = Typeconst "privkey"
		      then
			SCall(label,"makekey",[alg_arg_typed;pkey_arg_typed])
		      else raise
			(Exc_scenario
			   ("second argument \"" ^
			      (string_of_expression false pkey_arg) ^
			      "\" of makekey has type " ^
			      (string_of_type_expression pkey_arg_type) ^
			      " instead of privkey"))
		  | Typeconst "symalgo" ->
		      if pkey_arg_type = Typeconst "symkey"
		      then
			SCall(label,"makekey",[alg_arg_typed;pkey_arg_typed])
		      else raise
			(Exc_scenario
			   ("second argument \"" ^
			      (string_of_expression false pkey_arg) ^
			      "\" of makekey has type " ^
			      (string_of_type_expression pkey_arg_type) ^
			      " instead of symkey"))
		  | _ -> raise
		      (Exc_scenario
			 ("first argument \"" ^
			    (string_of_expression false alg_arg) ^
			    "\" of makekey has type " ^
			    (string_of_type_expression alg_arg_type) ^
			    " instead of algo or symalgo")))

	 | _ -> raise (Exc_scenario
			 ("makekey needs 2 arguments but is called with " ^
			  (string_of_int (List.length arguments)) ^
			  " arguments")))

  | SCall(label,rolename,arguments) ->
      (* at this point we know that rolename is not "makekey", hence *)
      (* we really have an invocation of a role here.                *)
      (try
	 let form_parameters = List.assoc rolename role_decl
	 in
	 let formal_length = List.length form_parameters
	 and actual_length = List.length arguments
	 in
	   if formal_length <> actual_length
	   then raise
	     (Exc_scenario
		("\"" ^ rolename ^ "\" is declared with " ^
		   (string_of_int formal_length) ^
		   " formal parameters but is here instantiated with " ^
		   (string_of_int actual_length) ^
		   " actual arguments"));
	   
	   let arguments_typed = 
	     try
	       List.map (type_expr sign env gl_var_decl []) arguments
	     with
		 Exc_typeexpression s -> raise (Exc_scenario s)
		 
	   in
	     List.iter2 
	       (fun (formname, (_,formtype)) expr_typed ->
		  let acttype = extract_type_expr expr_typed
		  in
		    if not (smaller_type acttype formtype)
		    then raise
		      (Exc_scenario
			 ("instantiaton of role \"" ^ rolename ^
			    "\": formal parameter \"" ^ formname ^
			    "\" is declared with type " ^
			    (string_of_type_expression formtype) ^
			    " but is instantiated with the expression \""^
			    (string_of_expression false expr_typed) ^
			    "\" which is of type "^
			    (string_of_type_expression  acttype))))
	       form_parameters
	       arguments_typed;

	     SCall(label, rolename, arguments_typed)
       with
	   Not_found -> raise
	     (Exc_scenario
		("instantiaton of a role \"" ^ rolename ^
		   "\" which is not declared")))

  | SForever (label,insts)  ->
      SForever(label,
	       List.map (type_scenario sign role_decl gl_var_decl env) insts)

  | SParallel (label,insts)  ->
      SParallel(label,
		List.map (type_scenario sign role_decl gl_var_decl env) insts)

  | SSequence (label,insts)  ->
      SSequence(label,
		List.map (type_scenario sign role_decl gl_var_decl env) insts)

  | SBlock (label,decl_list,insts)  ->
      check_declarations decl_list;
      SBlock(label,
	     decl_list,
	     List.map
	       (type_scenario sign role_decl gl_var_decl (decl_list@env))
	       insts)
	
  | SForall(label, decl_list, interval_list, body) ->
      check_declarations decl_list;
      check_interval_list interval_list;
      if ( List.length interval_list > 0 &&
	    List.length decl_list <> List.length interval_list)
      then raise (Exc_scenario
		    ("the number of variables (" ^ 
		       (string_of_int (List.length decl_list)) ^
		       ") and of intervals (" ^
		       (string_of_int (List.length interval_list)) ^
		       ") in a forall do not correspond"));
      SForall(label, decl_list, interval_list,
	      type_scenario sign role_decl gl_var_decl (decl_list@env) body)

  | SExists(label, decl_list, body) ->
      check_declarations decl_list;
      SExists(label, decl_list,
	      type_scenario sign role_decl gl_var_decl (decl_list@env) body)

  | SConditional(label, cond, thenlist, elselist) ->
      let cond_typed =
      try
	type_expr sign env gl_var_decl [] cond
      with
	  Exc_typeexpression s -> raise
	    (Exc_scenario
	       ("condition \"" ^ (string_of_expression false cond) ^
		  "\" can not be typed because: " ^ s))
      in let cond_type = extract_type_expr cond_typed
      in if cond_type <> Typeconst "bool"
	then raise (Exc_scenario
		      ("conditional with guard \"" ^
			 (string_of_expression false cond) ^
			 "\" which has type " ^
			 (string_of_type_expression cond_type) ^
			 " instead of bool"));
	SConditional(label, cond_typed,
		     type_scenlist sign role_decl gl_var_decl env thenlist,
		     type_scenlist sign role_decl gl_var_decl env elselist)

  | SAssign(label, lexp, exp) ->
      if not (is_l_expression lexp)
      then raise (Exc_scenario
		    ("assignment to \"" ^ (string_of_expression false lexp) ^
		     "\" which is not an l-expression"));
      let lexp_typed = try
	type_expr sign env gl_var_decl [] lexp
      with
	  Exc_typeexpression s -> 
	    raise (Exc_scenario
		     ("expression \"" ^ 
			(string_of_expression false lexp) ^
			"\" cannot be typed because "^ s))
      and exp_typed = try
	type_expr sign env gl_var_decl [] exp 
      with
	  Exc_typeexpression s -> 
	    raise (Exc_scenario
		     ("expression \"" ^ 
			(string_of_expression false exp) ^
			"\" cannot be typed because "^ s))
      in
      let lexp_type = extract_type_expr lexp
      and exp_type = extract_type_expr exp
      in if smaller_type exp_type lexp_type
	then SAssign(label, lexp_typed, exp_typed)
	else raise (Exc_scenario
		      ("assignment of \"" ^ (string_of_expression false exp) ^
			 "\" of type " ^ (string_of_type_expression exp_type) ^
			 " to variable \"" ^
			 (string_of_expression false lexp) ^
			 "\" of type " ^
			 (string_of_type_expression lexp_type)))

  | SNew(label, lexp) ->
      if not (is_l_expression lexp)
      then raise (Exc_scenario
		    ("\"new\" on \"" ^ (string_of_expression false lexp) ^
		     "\" which is not an l-expression"));
      let lexp_typed = try
	type_expr sign env gl_var_decl [] lexp
      with
	  Exc_typeexpression s -> 
	    raise (Exc_scenario
		     ("expression \"" ^ 
			(string_of_expression false lexp) ^
			"\" cannot be typed because "^ s))
      in
      let lexp_type = extract_type_expr lexp_typed
      in
	if (lexp_type = Typeconst "nonce" 
	    || lexp_type = Typeconst "int" 
	    || lexp_type = Typeconst "message")
	then
	  SNew(label, lexp_typed)
	else raise (Exc_scenario
		      ("\"new\" instruction on expression \"" ^
			 (string_of_expression false lexp) ^
			 "\" which is of type " ^ 
			 (string_of_type_expression lexp_type) ^ 
			 " but must be one of \"nonce\", \"int\"" ^
			 " or \"message\""))


and type_scenlist sign role_decl gl_var_decl env =
  List.map (type_scenario sign role_decl gl_var_decl env)

