(**************************************************************************)
(*  ProuveParser: parser frontend for the PROUVÉ language                *)
(*                                                                        *)
(*  Copyright (C) 2004-2009 Ralf Treinen                                  *)
(*                                                                        *)
(*  This library is free software; you can redistribute it and/or         *)
(*  modify it under the terms of the GNU Lesser General Public            *)
(*  License as published by the Free Software Foundation; either          *)
(*  version 2.1 of the License, or (at your option) any later version.    *)
(*                                                                        *)
(*  This library is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *)
(*  Lesser General Public License for more details.                       *)
(*                                                                        *)
(*  You should have received a copy of the GNU Lesser General Public      *)
(*  License along with this library; if not, write to the Free Software   *)
(*  Foundation, Inc., 59 Temple Place, Suite 330,                         *)
(*  Boston, MA  02111-1307  USA                                           *)
(*                                                                        *)
(*  Ralf Treinen <treinen@pps.jussieu.fr>                                 *)
(*  Laboratoire Preuves, Programmes et Systèmes (PPS),                   *)
(*  Université Paris Diderot, Paris, France                              *)
(*  http://www.pps.jussieu.fr/~treinen                                    *)
(*                                                                        *)
(*  This software was mainly written while the author was with the        *)
(*  Laboratoire Spécification et Vérification (LSV),                    *)
(*  ENS de Cachan, Cachan, France                                         *)
(**************************************************************************)

type type_expression =
    Typeconst of string                              (* Type constant *)
  | Typevar   of string                              (* Type variable *)
  | Typeapply of string * type_expression list       (* Type constructor *)
  | Typearrow of type_expression * type_expression   (* Function type *)
  | Typetuple of type_expression list                (* Tuple type *)
  | Typeundef

(* test whether a type is simple, that is only built of type constants, *)
(* tupling, and arrows.                                                 *)
let rec is_simple_type = function
    Typeconst _ -> true
  | Typetuple l -> List.for_all is_simple_type l
  | Typearrow(t1,t2) -> is_simple_type t1 && is_simple_type t2
  | _ -> false



(* the predefined type constants *)
let predefined_type_constants =
  ["int"; "bool"; "nonce";
   "algo"; "symalgo"; "pubkey"; "privkey"; "symkey";
   "principal"; "message"];;
(* the predefined type constructors with their arity *)
let predefined_type_constructors =
  [("list",1); ("table",2)];;

(********* Conversion to string ***************************************)

let rec string_of_type_expression_list = function 
    [] -> ""
  | h::r ->
      List.fold_left
	(fun s e -> s^", "^(string_of_type_expression e) )
	(string_of_type_expression h)
	r
and string_of_type_expression = function
    Typeconst s -> s
  | Typevar s -> s
  | Typeapply(f,l) -> f^"("^(string_of_type_expression_list l)^")"
  | Typearrow(from,res) ->
      (string_of_type_expression from) ^ " -> " ^
      (string_of_type_expression res)
  | Typetuple l -> "("^(string_of_type_expression_list l)^")"
  | Typeundef -> "#undef"

let string_of_nonempty_string_list = function
    [] -> failwith "This cannot happen"
  | h::r -> List.fold_left (fun s e -> s^", "^e ) h r


(********* Checking well-formedness *************************************)



(* calculate the free variables of a type expression *)
let rec free_variables = function
    Typeconst _ | Typeundef -> []
  | Typevar s -> [s]
  | Typeapply(_,l) | Typetuple l ->
      List.fold_left
	(fun frees exp -> List.merge compare frees (free_variables exp))
	[]
	l
  | Typearrow(from,res) ->
      Utils.listminus (free_variables res) (free_variables from) 

(* check whether a type expression is ground *)
let is_ground t = (List.length (free_variables t) = 0)

(* check that a type expression does not contain an arrow *)
let rec arrow_free = function
    Typeconst _ | Typevar _ | Typeundef  -> true
  | Typearrow _ -> false
  | Typeapply(_,l) | Typetuple l -> List.for_all arrow_free l

(* check that a type is either arrow-free, or of the form (t1 -> t2) *)
(* where t1 and t2 are arrow-free                                    *)
let atmost_top_arrow t =
  arrow_free t ||
  (match t with
       Typearrow(t1,t2) -> arrow_free t1 && arrow_free t2
     | _ -> false
  )
(* check that only predefined type constants, and predefined type    *)
(* constructors are used with the right arity                        *)
exception Exc_typecons of string
let rec check_type_constructors = function
    Typevar _ | Typeundef  -> ()
  | Typetuple l -> List.iter check_type_constructors l
  | Typearrow(t1,t2) ->
      check_type_constructors t1;
      check_type_constructors t2
  | Typeconst s ->
      if not (List.mem s predefined_type_constants)
      then raise (Exc_typecons (" uses a type constant \"" ^ s ^
				"\" which is not defined.")) 
  | Typeapply(f,l) ->
      try
	let arity = List.assoc f predefined_type_constructors
	and used_length = List.length l
	in
	  if used_length <> arity
	  then raise (Exc_typecons
			(" uses the type constructor \"" ^ f ^
			 "\" with arity " ^ (string_of_int used_length)  ^
			 " instead of " ^ (string_of_int arity) ^ "."))
      with
	  Not_found -> raise (Exc_typecons
				(" uses a type constructor \"" ^ f ^
				 "\" which is not defined.")) 


(* check that a type expression is well-formed *)
exception Wellformed_type of string                            
let well_formed t = begin
  let fv = free_variables t
  in

    (* check that type expression is ground *)
    if List.length fv > 0
    then raise (Wellformed_type
		  ("the type expression \"" ^ 
		   (string_of_type_expression t) ^
		   "\" contains free variables: " ^
		   (string_of_nonempty_string_list fv) ^ "."));

    (* check that arrow-free, or an arrow of two arrow-free type *)
    (* expressions.                                              *)
    if not (atmost_top_arrow t)
    then raise (Wellformed_type
		  ("the type expression \"" ^ 
		   (string_of_type_expression t) ^
		   "must be either arrow-free, or of the form t1->t2" ^
		   "where t1 and t2 are arrow-free."));

    (* check type constants and type constructors *)
    try check_type_constructors t
    with Exc_typecons s ->
      raise (Wellformed_type
	       ("the type expression \"" ^ 
		(string_of_type_expression t) ^ "\"" ^ s))
end

(****** Subtyping *****************************************************)

(* (sup_type t1 t2) returns the supremum of two ground type expressions *)
(* in the flat cpo with bottom element Typeundef and top tMessage       *)
let tMessage = Typeconst "message"
let rec sup_type t1 t2 =
  assert (is_ground t1);
  assert (is_ground t2);
  match t2 with
    Typeundef -> t1
  | Typevar x2 -> (match t1 with
		      Typevar x1 ->
			if x1 = x2
			then t1
			else tMessage
		    | Typeundef -> t2
		    | _ -> tMessage) 
  | Typeconst x2 -> (match t1 with
			Typeconst x1 ->
			  if x1 = x2
			  then t1
			  else tMessage
		      | Typeundef -> t2
		      | _ -> tMessage)
  | Typetuple l2 -> (match t1 with
			 Typetuple l1 ->
			   (try
			     Typetuple (List.map2 sup_type l1 l2)
			   with
			       Invalid_argument _ -> tMessage)
		       | Typeundef -> t2
		       | _ -> tMessage)
  | Typeapply(f2,l2) -> (match t1 with
			     Typeapply(f1,l1) ->
			       if f1=f2
			       then
				 (try 
				    Typeapply(f1, List.map2 sup_type l1 l2)
				  with
				      Invalid_argument _ -> tMessage)  
			       else tMessage
			   | Typeundef -> t2
			   | _ -> tMessage)
  | Typearrow(from2,res2) -> 
      failwith "this cannot happen: sup with an arrow type"

(* (has_inf_type t1 t2) tests whether the type expressions t1 and t2 *)
(* have an infimum.                                                  *)
let rec has_inf_type t1 t2 =
(*  assert (is_ground t1 && arrow_free t1 && t1 <> Typeundef);
  assert (is_ground t2 && arrow_free t2 && t2 <> Typeundef); *)
  match t2 with
      Typeconst s2 -> 
	(match t1 with
	     Typeconst s1 -> s1=s2 || s1="message" || s2 = "message"
	   | _ -> false)
    | Typetuple l2 ->
	(match t1 with
	     Typetuple l1 -> List.for_all2 has_inf_type l1 l2
	   | Typeconst "message" -> true
	   | _ -> false)
    | Typeapply(f2,l2) ->
	(match t1 with
	     Typeapply(f1,l1) -> f1 = f2 &&
				      List.for_all2 has_inf_type_undef l1 l2
	   | Typeconst "message" -> true
	   | _ -> false)
    | Typevar _ -> failwith "this cannot happen: variable in has_inf_type"
    | Typeundef -> failwith "this cannot happen: undef in has_inf_type"
    | Typearrow _ -> failwith "this cannot happen: arrow in has_inf_type"
and has_inf_type_undef t1 t2 =
	      t1 = Typeundef || t2=Typeundef || has_inf_type t1 t2


(* (smaller_type t1 t2), where t1 and t2 are ground and arrow_free, tests *)
(* whether t1 is smaller than t2.                                         *)
let rec smaller_type t1 t2 =
  assert (is_ground t1 && arrow_free t1);
  assert (is_ground t2 && arrow_free t2);
  t1 = Typeundef || match t2 with
    | Typeconst "message" -> true
    | Typeconst s2 ->
	(match t1 with
	     Typeconst s1 -> s1=s2
	   | _ -> false)
    | Typetuple l2 ->
	(match t1 with
	     Typetuple l1 ->
	       (List.length l1) = (List.length l2) &&
	    (List.for_all2 smaller_type l1 l2)
	   | _ -> false)
    | Typeapply (f2,l2) -> 
	(match t1 with
	     Typeapply(f1,l1) -> (f1 = f2) &&
	       (List.length l1) = (List.length l2)
	       && List.for_all2 smaller_type l1 l2
	   | _ -> false)
    | Typeundef -> 
	(match t1 with
	     Typeundef -> true
	   | _ -> false)
    | Typevar _ -> failwith "this cannot happen: smaller_type with var"
    | Typearrow _ -> failwith "this cannot happen: smaller_type with arrow"
			


(***** Substitutions ******************************************************)

(* a substitution is an association list that associates type expressions *)
(* to strings. No string is declared twice in a substitution.             *)
exception Nomatch_type 

(* (merge_subst subs1 subs2) merges the two substitution, or raises an *)
(* exception in case of a conflict.                                    *)
let rec merge_subst subs = function
    [] -> subs
  | (s,t)::r ->
      (try
	(s, (sup_type t (List.assoc s subs)))
      with
	  Not_found -> (s,t))
      ::(merge_subst subs r)


	    
(* (merge_substlist l), where l is a a list of substitutions, merges all *)
(* the substitutions in l (or raises an exception in case of a conflict. *)
let merge_substlist l = List.fold_left merge_subst [] l

let rec type_match pattern gtype = 
  (* the pattern cannot contain Typeundef *)
  assert (arrow_free pattern);
  assert (arrow_free gtype);
  assert (is_ground gtype);
  match pattern with
      Typevar s -> [(s,gtype)]
    | Typeundef -> failwith "this cannot happen: #undef in type pattern"
    | Typeconst s ->
	if s = "message" || pattern = gtype
	then []
	else raise Nomatch_type
    | Typetuple lp ->
	(match gtype with
	     Typetuple lt ->
	       if (List.length lp) = (List.length lt)
	       then merge_substlist (List.map2 type_match lp lt)
	       else raise Nomatch_type
	   | _ -> raise Nomatch_type)
    | Typeapply(fp,lp) ->
	(match gtype with
	     Typeapply(ft,lt) ->
	       if fp=ft && (List.length lp) = (List.length lt)
	       then merge_substlist (List.map2 type_match lp lt)
	       else raise Nomatch_type
	   | _ -> raise Nomatch_type)
    | Typearrow(from,res) ->
	(match gtype with
	     Typearrow(gfrom,gres) ->
	       merge_subst (type_match from gfrom) (type_match res gres)
	   | _ -> raise Nomatch_type)

let rec apply_type_subst subst = function
    Typevar s as t -> (try List.assoc s subst with Not_found -> t)
  | Typeundef -> Typeundef
  | Typeconst _  as t -> t
  | Typetuple l -> Typetuple (List.map (apply_type_subst subst) l)
  | Typeapply(f,l) ->
      Typeapply(f, List.map (apply_type_subst subst) l)
  | Typearrow(from,res) -> 
      Typearrow (apply_type_subst subst from, apply_type_subst subst res)
     



